@@ -1,33 +1,62 @@
 //! Library for strings of fixed maximum lengths that can be copied and
-//! stack-allocated using Rust's new const generics feature.
+//! stack-allocated using const generics.
 //!
-//! **The main structure provided by this crate are [fstr] and [zstr].**
+//! **The structures provided by this crate are [fstr], [zstr]** and tstr.
+//! However, tstr is not exported and can only be used through the type
+//! aliases [str8], [str16], [str32], through [str256].
 //!
-//! **Version 0.2.x** adds **unicode support** and a new module for
+//! **Version 0.2.x** adds **unicode support** and a module for
 //! **zero_terminated strings** in the structure [zstr].
 //! These strings are more memory efficient than [fstr] but less efficient
 //! in terms of run time.
-
-//! Example:
+//! 
+//! For version 0.2.2, the [str8] through [str256] type aliases where
+//! changed to refer to another, internal type distinct from [fstr] and
+//! [zstr].  This type represents strings of up to 255 bytes with a
+//! `[u8;N]` underneath where it's assumed that N<=256.  The first
+//! byte of the array holds the length of the string in bytes.  This structure
+//! represents the best combination of fstr and zstr in terms of speed
+//! and memory efficiency.  However, because Rust does not currently provide
+//! a why to specify conditions on const generics at compile time, such as
+//! `where N<=256`, the internal "tiny string" type *tstr* is not exported and can
+//! only be used through the aliases.  These strings implement that same
+//! functions and traits as [fstr] and [zstr] so the documentation for
+//! these structures also apply to the hidden type.
 //!
-//!```ignore
-//!   let s1 = str16::from("abc"); // str16 is alias for fstr<16>
-//!   let mut s2 = str32::from("and xyz");
-//!   s2.push(" and 1234");  // adds to end of s2
-//!   println!("{} {}, {}", s1, &s2, s2.len());  
-//!   println!("{}", &s1=="abc");   // can compare with &str
-//!   let s3 = s1;     // copied, not moved
-//!   println!("{}", "abc"==&s1);
-//!   println!("{}, {} ", s1==s3, s1==s2.resize()); 
-//!   println!("{}", s2.substr(2,6));
-//!   let s4:fstr<64> = s1.resize();  // resize copies to new-capacity fstr
-//!   let owned_string = s4.to_string();
-//!   let str_slice:&str = s4.to_str();
-//!   let z:zstr<8> = zstr::from("λxλy.x");
+//! For version 0.2.2  the fsiter construct and direct iterator
+//! implmentation for fstr has been removed. Use the [fstr::chars]
+//! function instead.
+
+//!  ## Examples
+//!
 //!```
+//! let a:fstr<8> = fstr::from("abcdefg"); //creates fstr from &str
+//! let a1:fstr<8> = a; // copied, not moved
+//! let a2:&str = a.to_str();
+//! let a3:String = a.to_string();
+//! assert_eq!(a.nth_ascii(2), 'c');
+//! let ab = a.substr(1,5);  // copies substring to new fstr
+//! assert_eq!(ab,"bcde");  // can compare with &str
+//! assert!(a<ab);  // implements Ord trait (and Hash, Debug, Display)
+//! let mut u:fstr<8> = fstr::from("aλb"); //unicode support
+//! for x in u.nth(1) {assert_eq!(x,'λ');} // nth returns Option<char>
+//! assert!(u.set(1,'μ'));  // changes a character of the same character class
+//! assert!(!u.set(1,'c')); // .set returns false on failure
+//! assert!(u.set(2,'c'));
+//! assert_eq!(u, "aμc");
+//! assert_eq!(u.len(),4);  // length in bytes
+//! assert_eq!(u.charlen(),3);  // length in chars
+//! let mut ac:fstr<16> = a.resize(); // copies to larger capacity string
+//! let remainder:&str = ac.push("hijklmnopqrst");  //appends string, returns left over
+//! assert_eq!(ac.len(),16);
+//! assert_eq!(remainder, "qrst");
+//! ac.truncate(10); // shortens string in place
+//! assert_eq!(&ac,"abcdefghij");
+//!```
+//!
+//![zstr] and the type aliases [str8]...[str256] implement the same capabilities as [fstr].
+
 
-// Fixed, :Copy strings of limited size.  size of each fstr is N or less,
-// array chrs is 0-terminated if size of string is less than N:
 
 #![allow(unused_variables)]
 #![allow(non_snake_case)]
@@ -39,10 +68,12 @@
 
 pub mod zero_terminated;
 pub use zero_terminated::*;
+mod tiny_internal;
+use tiny_internal::*;
 
 use std::cmp::Ordering;
 
-/// main type: fixed string of size up to N:
+/// main type: string of size up to const N:
 #[derive(Copy,Clone,Eq,PartialEq,Hash)]
 pub struct fstr<const N:usize>
 {
@@ -56,7 +87,7 @@ impl<const N:usize> fstr<N>
   /// several others including [fstr::from].
   pub fn make(s:&str) -> fstr<N>
    {
-      if (N>65536 || N<1) {panic!("Valid fstr strings are limited to fstr<1> to zstr<65536>");}
+//      if (N>65536 || N<1) {panic!("Valid fstr strings are limited to fstr<1> to zstr<65536>");}
       let bytes = s.as_bytes(); // &[u8]
       let blen = bytes.len();
       let mut chars = [0u8; N];
@@ -93,6 +124,9 @@ impl<const N:usize> fstr<N>
    /// with [str::len]
    pub fn len(&self)->usize { self.len }
 
+   /// returns maximum capacity in bytes
+   pub fn capacity(&self) -> usize { N }
+
    /// converts fstr to an owned string
    pub fn to_string(&self) -> String
    {
@@ -181,12 +215,27 @@ impl<const N:usize> fstr<N>
       self.to_str().chars().nth(n)   
    }
 
+   /// returns the nth byte of the string as a char.  This
+   /// function should only be called on ascii strings.  It
+   /// is designed to be quicker than [fstr::nth], and does not check array bounds or
+   /// check n against the length of the string. Nor does it check
+   /// if the value returned is within the ascii range.
+   pub fn nth_ascii(&self,n:usize) -> char
+   {
+      self.chrs[n] as char
+   }
+
+   /// determines if string is an ascii string
+   pub fn is_ascii(&self) -> bool
+   { self.to_str().is_ascii() }
+   
    /// shortens the fstr in-place (mutates).  If n is greater than the
-   /// current length of the string, this operation will have no effect.
+   /// current length of the string in chars, this operation will have no effect.
    pub fn truncate(&mut self, n:usize)
    {
      if let Some((bi,c)) = self.to_str().char_indices().nth(n) {
-        self.chrs[bi] = 0;
+        //self.chrs[bi] = 0;
+        self.len = bi;
      }
      //if n<self.len {self.len = n;}
    }
@@ -259,6 +308,21 @@ impl<const N:usize,const M:usize> std::convert::From<&zstr<M>> for fstr<N>
   }
 }
 
+impl<const N:usize,const M:usize> std::convert::From<tstr<M>> for fstr<N>
+{
+  fn from(s:tstr<M>) -> fstr<N>
+  {
+     fstr::<N>::make(&s.to_str())
+  }
+}
+impl<const N:usize,const M:usize> std::convert::From<&tstr<M>> for fstr<N>
+{
+  fn from(s:&tstr<M>) -> fstr<N>
+  {
+     fstr::<N>::make(&s.to_str())
+  }
+}
+
 impl<const N:usize> std::cmp::PartialOrd for fstr<N>
 {
    fn partial_cmp(&self, other:&Self) -> Option<Ordering>
@@ -321,7 +385,7 @@ impl<const N:usize> std::convert::From<&str> for fstr<N>
   }
 }
 */
-
+/*
 /// [IntoIterator] struct for fstr
 pub struct fstriter<const N:usize>
 {
@@ -351,6 +415,7 @@ impl<const N:usize> IntoIterator for fstr<N>
      }
   }
 }//IntoIterator
+*/
 
 impl<const N:usize> std::fmt::Display for fstr<N>
 {
@@ -360,11 +425,12 @@ impl<const N:usize> std::fmt::Display for fstr<N>
   }
 }
 
+
 impl<const N:usize> PartialEq<&str> for fstr<N>
 {
   fn eq(&self, other:&&str) -> bool
   {
-     self==other   // see below
+     &self.to_str()==other   // see below
   }//eq
 }
 impl<const N:usize> PartialEq<&str> for &fstr<N>
@@ -382,7 +448,7 @@ impl<'t, const N:usize> PartialEq<fstr<N>> for &'t str
 impl<'t, const N:usize> PartialEq<&fstr<N>> for &'t str
 {
   fn eq(&self, other:&&fstr<N>) -> bool
-  { other==self }
+  { &other.to_str()==self }
 }
 
 /// defaults to empty string
@@ -460,9 +526,31 @@ impl<const N:usize> fstr<N>
 
 
 
-/// types for small strings 
-pub type str8 = fstr<8>; 
-pub type str16 = fstr<16>; 
-pub type str32 = fstr<32>; 
-pub type str64 = fstr<64>; 
-pub type str128 = fstr<128>;
+/// types for small strings that use a more efficient representation
+/// underneath.  A str8 can hold a string of up to 7 bytes (7 ascii chars).
+/// The same functions for [fstr] and [zstr] are provided for these types.
+///
+/// Example:
+/// ```
+///  let mut s = str8::from("aλc");
+///  assert_eq!(s.capacity(),7);
+///  assert_eq!(s.push("1234567"), "4567");
+///  assert_eq!(s,"aλc123");
+///  assert_eq!(s.charlen(), 6);
+///  assert_eq!(s.len(), 7);  
+/// ```
+pub type str8 = tstr<8>;
+/// A str16 can hold a string of up to 15 bytes. See docs for [fstr] or [zstr]
+pub type str16 = tstr<16>;
+/// A str32 can hold a string of up to 31 bytes. See docs for [fstr] or [zstr]
+pub type str32 = tstr<32>;
+/// A str64 can hold a string of up to 63 bytes. See docs for [fstr] or [zstr]
+pub type str64 = tstr<64>;
+/// A str28 can hold a string of up to 127 bytes. See docs for [fstr] or [zstr]
+pub type str128 = tstr<128>;
+/// Each type strN is represented underneath by a `[u8;N]` with N<=256.
+/// The first byte of the array always holds the length of the string.
+/// Each such type can hold a string of up to N-1 bytes, with max size=255.
+/// These types represent the best compromise between [fstr] and [zstr] in
+/// terms of speed and memory efficiency.
+pub type str256 = tstr<256>;
